"""Manifest declaration for the exaSPIM capsules"""
import os
import sys
from datetime import datetime
from typing import Optional, Tuple

from aind_data_schema import DataProcess, Processing
from aind_data_schema.base import AindModel
from aind_data_schema.data_description import Institution
from aind_data_transfer.util import file_utils
from pydantic import Field


# Based on aind-data-transfer/scripts/processing_manifest.py


class DatasetStatus(AindModel):  # pragma: no cover
    """Status of the datasets to control next processing step. TBD"""

    # status: Status = Field(
    #     ...,
    #     description="Status of the dataset on the local storage",
    #     title="Institution",
    #     enumNames=[i.value for i in Status],
    # )
    # Creating datetime
    status_date = Field(
        datetime.now().date().strftime("%Y-%m-%d"),
        title="Date the flag was created",
    )
    status_time = Field(
        datetime.now().time().strftime("%H-%M-%S"),
        title="Time the flag was created",
    )


class N5toZarrParameters(AindModel):  # pragma: no cover
    """N5 to zarr conversion configuration parameters.

    n5tozarr_da_converter Code Ocean task config parameters."""

    voxel_size_zyx: Tuple[float, float, float] = Field(
        ..., title="Z,Y,X voxel size in micrometers for output metadata"
    )

    input_uri: str = Field(
        ...,
        title="Input N5 dataset path. Must be a local filesystem path or "
        "start with s3:// to trigger S3 direct access.",
    )

    output_uri: str = Field(
        ...,
        title="Output Zarr dataset path. Must be a local filesystem path or "
        "start with s3:// to trigger S3 direct access. "
        "Must be different from the input_uri. Will be overwritten if exists.",
    )


class ZarrMultiscaleParameters(AindModel):  # pragma: no cover
    """N5 to zarr conversion configuration parameters.

    zarr_multiscale Code Ocean task config parameters."""

    voxel_size_zyx: Tuple[float, float, float] = Field(
        ..., title="Z,Y,X voxel size in micrometers for output metadata"
    )

    input_uri: str = Field(
        ...,
        title="Input Zarr group dataset path. Must be a local filesystem path or "
        "start with s3:// to trigger S3 direct access.",
    )

    output_uri: Optional[str] = Field(None, title="Output Zarr group dataset path if different from input.")


class ExaspimProcessingPipeline(AindModel):  # pragma: no cover
    """ExaSPIM processing pipeline configuration parameters

    If a field is None, it is considered to be a disabled step."""

    n5_to_zarr: N5toZarrParameters = Field(None, title="N5 to single scale Zarr conversion")
    zarr_multiscale: ZarrMultiscaleParameters = Field(None, title="Zarr to multiscale Zarr conversion")


class ExaspimManifest(AindModel):  # pragma: no cover
    """Manifest definition of an exaSPIM processing session.

    Connects the dataset and its pipeline processing history."""

    schema_version: str = Field("0.1.0", title="Schema Version", const=True)
    license: str = Field("CC-BY-4.0", title="License", const=True)

    specimen_id: str = Field(..., title="Specimen ID")
    # dataset_status: DatasetStatus = Field(
    #     ..., title="Dataset status", description="Dataset status"
    # )
    institution: Institution = Field(
        ...,
        description="An established society, corporation, foundation or other organization "
        "that collected this data",
        title="Institution",
        enumNames=[i.value.name for i in Institution],
    )
    # acquisition: Acquisition = Field(
    #     ...,
    #     title="Acquisition data",
    #     description="Acquition data coming from the rig which is necessary to create matadata files",
    # )

    processing_pipeline: ExaspimProcessingPipeline = Field(
        ...,
        title="ExaSPIM pipeline parameters",
        description="Parameters necessary for the exaspim pipeline steps.",
    )
    processing: Processing = Field(
        ...,
        title="ExaSPIM pipeline processing steps log",
        description="Processing steps that has already taken place.",
    )


def create_example_manifest(printit=False) -> ExaspimManifest:  # pragma: no cover
    """Create example manifest file

    Parameters
    ----------
    printit: bool
      Print the example?

    Returns
    -------
    example_manifest: ExaspimManifest
    """
    # print(ProcessingManifest.schema_json(indent=2))
    # print(ProcessingManifest.schema())

    processing_manifest_example = ExaspimManifest(
        specimen_id="653431",
        institution=Institution.AIND,
        processing_pipeline=ExaspimProcessingPipeline(
            n5_to_zarr=N5toZarrParameters(
                voxel_size_zyx=(1.0, 0.748, 0.748),
                input_uri="s3://aind-scratch-data/gabor.kovacs/2023-07-25_1653_BSS_fusion_653431/ch561/",
                output_uri="s3://aind-scratch-data/gabor.kovacs/n5_to_zarr_CO_2023-08-17_1351/",
            ),
            zarr_multiscale=ZarrMultiscaleParameters(
                voxel_size_zyx=(1.0, 0.748, 0.748),
                input_uri="s3://aind-scratch-data/gabor.kovacs/2023-07-25_1653_BSS_fusion_653431/ch561/",
            ),
        ),
        processing=Processing(data_processes=[]),
    )

    if printit:
        print(processing_manifest_example.json(indent=3))
    return processing_manifest_example


def get_capsule_manifest():  # pragma: no cover
    """Get the manifest file from its Code Ocean location or as given in the cmd-line argument.

    Raises
    ------
    If the manifest is not found, required fields will be missing at schema validation.
    """
    if len(sys.argv) > 1:
        manifest_name = sys.argv[1]
    else:
        manifest_name = "data/manifest/exaspim_manifest.json"
    json_data = file_utils.read_json_as_dict(manifest_name)
    return ExaspimManifest(**json_data)


def append_metadata_to_manifest(capsule_manifest: ExaspimManifest, process: DataProcess) -> None:
    """Append the given dataprocess metadata to the exaspim pipeline manifest

    So long the pipeline is a linear sequence of steps, this should always be the
    case.
    """
    capsule_manifest.processing.data_processes.append(process)


def write_result_manifest(capsule_manifest: ExaspimManifest) -> None:
    """Write the updated manifest file to the Code Ocean results folder."""
    os.makedirs("results/manifest", exist_ok=True)
    with open("results/manifest/exaspim_manifest.json", "w") as f:
        f.write(capsule_manifest.json(indent=3))


def write_result_metadata(capsule_metadata: DataProcess) -> None:
    """Write the metadata file to the Code Ocean results folder."""
    os.makedirs("results/meta", exist_ok=True)
    with open("results/meta/exaspim_process.json", "w") as f:
        f.write(capsule_metadata.json(indent=3))


if __name__ == "__main__":
    create_example_manifest(printit=True)



"""Wrapper functions and runtime arguments definition."""
import json
import logging
import os
import selectors
import shutil
import subprocess
import sys
from typing import Dict, List, Union

import argschema
import argschema.fields as fld
import marshmallow as mm
import psutil

from .imagej_macros import ImagejMacros


class IPDetectionSchema(argschema.ArgSchema):  # pragma: no cover
    """Adjustable parameters to detect IP."""

    downsample = fld.Int(
        required=True,
        metadata={"description": "Downsampling factor. Use the one that is available in the dataset."},
    )
    bead_choice = fld.String(
        required=True,
        validate=mm.validate.OneOf(list(ImagejMacros.MAP_BEAD_CHOICE.keys())),
        metadata={"description": "Beads detection mode"},
    )
    sigma = fld.Float(
        load_default=1.8, metadata={"description": "Difference of Gaussians sigma (beads_mode==manual only)."}
    )
    threshold = fld.Float(
        load_default=0.1,
        metadata={"description": "Difference of Gaussians detection threshold (beads_mode==manual only)."},
    )
    find_minima = fld.Boolean(
        load_default=False, metadata={"description": "Find minima (beads_mode==manual only)."}
    )
    find_maxima = fld.Boolean(
        load_default=True, metadata={"description": "Find maxima (beads_mode==manual only)."}
    )
    set_minimum_maximum = fld.Boolean(
        load_default=False,
        metadata={"description": "Define the minimum and maximum intensity range manually"},
    )
    minimal_intensity = fld.Float(
        load_default=0, metadata={"description": "Minimal intensity value (if set_minimum_maximum==True)."}
    )
    maximal_intensity = fld.Float(
        load_default=65535,
        metadata={"description": "Minimal intensity value (if set_minimum_maximum==True)."},
    )
    maximum_number_of_detections = fld.Int(
        load_default=0,
        metadata={
            "description": "If not equal to 0, the number of maximum IPs to detect."
            " Set ip_limitation_choice, too."
        },
    )
    ip_limitation_choice = fld.String(
        required=True,
        validate=mm.validate.OneOf(list(ImagejMacros.MAP_IP_LIMITATION_CHOICE.keys())),
        metadata={
            "description": "How to pick limit_amount_of_detections is set >0 and the maximum number is hit."
        },
    )


class IPRegistrationSchema(argschema.ArgSchema):  # pragma: no cover
    """Adjustable parameters to register with translation only."""

    transformation_choice = fld.String(
        required=True,
        validate=mm.validate.OneOf(list(ImagejMacros.MAP_TRANSFORMATION.keys())),
        metadata={"description": "Translation, rigid or full affine transformation ?"},
    )

    compare_views_choice = fld.String(
        required=True,
        validate=mm.validate.OneOf(list(ImagejMacros.MAP_COMPARE_VIEWS.keys())),
        metadata={"description": "Which views to compare ?"},
    )

    interest_point_inclusion_choice = fld.String(
        required=True,
        validate=mm.validate.OneOf(list(ImagejMacros.MAP_INTEREST_POINT_INCLUSION.keys())),
        metadata={"description": "Which interest points to use ?"},
    )

    fix_views_choice = fld.String(
        required=True,
        validate=mm.validate.OneOf(list(ImagejMacros.MAP_FIX_VIEWS.keys())),
        metadata={"description": "Which views to fix ?"},
    )

    fixed_tile_ids = fld.List(
        fld.Int,
        load_default=[
            0,
        ],
        metadata={"description": "Setup ids of fixed tiles (fix_views_choice==select_fixed)."},
    )
    map_back_views_choice = fld.String(
        required=True,
        validate=mm.validate.OneOf(list(ImagejMacros.MAP_MAP_BACK_VIEWS.keys())),
        metadata={"description": "How to map back views?"},
    )
    map_back_reference_view = fld.Int(
        load_default=0, metadata={"description": "Selected reference view for map back."}
    )
    do_regularize = fld.Boolean(default=False, metadata={"description": "Do regularize transformation?"})
    regularization_lambda = fld.Float(
        load_default=0.1, metadata={"description": "Regularization lambda (do_regularize==True only)."}
    )
    regularize_with_choice = fld.String(
        load_default="rigid",
        validate=mm.validate.OneOf(list(ImagejMacros.MAP_REGULARIZATION.keys())),
        metadata={"description": "Which regularization to use (do_regularize==True only) ?"},
    )


class ImageJWrapperSchema(argschema.ArgSchema):  # pragma: no cover
    """Command line arguments."""

    session_id = fld.String(required=True, metadata={"description": "Processing run session identifier"})
    memgb = fld.Int(
        required=True,
        metadata={
            "description": "Allowed Java interpreter memory. "
            "Should be about 0.8 GB x number of parallel threads less than total available."
        },
    )
    parallel = fld.Int(
        required=True,
        metadata={"description": "Number of parallel Java worker threads."},
        validate=mm.validate.Range(min=1, max=128),
    )
    dataset_xml = fld.String(required=True, metadata={"description": "Input xml dataset definition"})
    do_detection = fld.Boolean(required=True, metadata={"description": "Do interest point detection?"})
    ip_detection_params = fld.Nested(
        IPDetectionSchema, required=False, metadata={"description": "Interest point detection parameters"}
    )
    do_registrations = fld.Boolean(
        required=True,
        metadata={"description": "Do first transformation fitting ?"},
    )
    ip_registrations_params = fld.Nested(
        IPRegistrationSchema,
        required=False,
        metadata={"description": "Registration parameters (do_registrations==True only)"},
        many=True,
    )


def wrapper_cmd_run(cmd: Union[str, List], logger: logging.Logger) -> int:
    """Wrapper for a shell command.

    Wraps a shell command.

    It monitors, captures and re-prints stdout and strderr as the command progresses.

    TBD: Validate the program output on-the-fly and kill it if failure detected.

    Parameters
    ----------
    cmd: `str`
        Command that we want to execute.

    logger: `logging.Logger`
        Logger instance to use.

    Returns
    -------
    r: `int`
      Cmd return code.
    """
    logger.info("Starting command (%s)", str(cmd))
    p = subprocess.Popen(cmd, bufsize=128, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    sel = selectors.DefaultSelector()
    try:
        sel.register(p.stdout, selectors.EVENT_READ)
        sel.register(p.stderr, selectors.EVENT_READ)
        while p.poll() is None:  # pragma: no cover
            for key, _ in sel.select():
                data = key.fileobj.read1().decode()
                if not data:
                    continue
                if key.fileobj is p.stdout:
                    print(data, end="")
                else:
                    print(data, end="", file=sys.stderr)
        # Ensure to process everything that may be left in the buffer
        data = p.stdout.read().decode()
        if data:
            print(data, end="")
        data = p.stderr.read().decode()
        if data:
            print(data, end="", file=sys.stderr)
    finally:
        p.stdout.close()
        p.stderr.close()
        sel.close()
    r = p.wait()
    logger.info("Command finished with return code %d", r)
    return r


def get_auto_parameters(args: Dict) -> Dict:
    """Determine environment parameters.

    Determine number of cpus, imagej memory limit and imagej macro file names.

    Note
    ----

    Number of cpus and memory are for the whole VM at the moment, not what is available
    for the capsule.

    Parameters
    ----------
    args: `Dict`
        ArgSchema args dictionary

    Returns
    -------
    params: `Dict`
      New dictionary with determined parameters.

    """
    ncpu = os.cpu_count()

    mem_GB = psutil.virtual_memory().total // (1024 * 1024 * 1024)
    d = int(mem_GB * 0.1)
    if d < 5:
        d = 5
    mem_GB -= d
    if mem_GB < 10:
        raise ValueError("Too little memory available")

    process_xml = "/results/bigstitcher_{session_id}.xml".format(**args)
    macro_ip_det = "/results/macro_ip_det_{session_id}.ijm".format(**args)
    return {
        "process_xml": process_xml,
        # Do not use, this is the whole VM at the moment, not what is available for the capsule
        "auto_ncpu": ncpu,
        # Do not use, this is the whole VM at the moment, not what is available for the capsule
        "auto_memgb": mem_GB,
        "macro_ip_det": macro_ip_det,
    }


def main():  # pragma: no cover
    """Entry point if run as a standalone program."""
    logging.basicConfig(format="%(asctime)s %(levelname)-7s %(message)s")

    logger = logging.getLogger()
    parser = argschema.ArgSchemaParser(schema_type=ImageJWrapperSchema)

    args = dict(parser.args)
    logger.setLevel(args["log_level"])
    args.update(get_auto_parameters(args))
    logger.info("Invocation: %s", sys.argv)

    logger.info("Writing out config.json")
    with open("/results/config.json", "w") as f:
        json.dump(args, f, indent=2)

    logger.info("Copying input xml %s -> %s", args["dataset_xml"], args["process_xml"])
    shutil.copy(args["dataset_xml"], args["process_xml"])

    if args["do_detection"]:
        det_params = dict(args["ip_detection_params"])
        det_params["parallel"] = args["parallel"]
        det_params["process_xml"] = args["process_xml"]
        logger.info("Creating macro %s", args["macro_ip_det"])
        with open(args["macro_ip_det"], "w") as f:
            f.write(ImagejMacros.get_macro_ip_det(det_params))
        r = wrapper_cmd_run(
            [
                "ImageJ",
                "-Dimagej.updater.disableAutocheck=true",
                "--headless",
                "--memory",
                "{memgb}G".format(**args),
                "--console",
                "--run",
                args["macro_ip_det"],
            ],
            logger,
        )
        if r != 0:
            raise RuntimeError("IP detection command failed.")
    else:
        if args["do_registrations"]:
            # We assume that interest point detections are already present in the input dataset
            # in the folder of the xml dataset file
            logger.info("Assume already detected interestpoints.")
            ip_src = os.path.join(os.path.dirname(args["dataset_xml"]), "interestpoints")
            logger.info("Copying %s -> /results/", ip_src)
            shutil.copytree(ip_src, "/results/interestpoints", dirs_exist_ok=True)

    if args["do_registrations"]:
        if "ip_registrations_params" not in args:
            raise ValueError("Registration steps are requested but no configuration provided.")
        reg_index = 0
        for reg_params in args["ip_registrations_params"]:
            macro_reg = f"/results/macro_ip_reg{reg_index:d}.ijm"
            reg_params = dict(reg_params)
            reg_params["process_xml"] = args["process_xml"]
            reg_params["parallel"] = args["parallel"]
            logger.info("Creating macro %s", macro_reg)
            with open(macro_reg, "w") as f:
                f.write(ImagejMacros.get_macro_ip_reg(reg_params))
            r = wrapper_cmd_run(
                [
                    "ImageJ",
                    "-Dimagej.updater.disableAutocheck=true",
                    "--headless",
                    "--memory",
                    "{memgb}G".format(**args),
                    "--console",
                    "--run",
                    macro_reg,
                ],
                logger,
            )
            if r != 0:
                raise RuntimeError("IP registration1 command failed.")
            reg_index += 1

    logger.info("Done.")


if __name__ == "__main__":  # pragma: no cover
    main()



"""ExaSPIM cloud conversion of N5 to multiscale ZARR using Dask.array"""
import datetime
import logging
import sys

import xarray_multiscale
from aind_data_schema import DataProcess
from aind_data_schema.processing import ProcessName
from aind_data_transfer.transformations.ome_zarr import _get_first_mipmap_level
from aind_data_transfer.util.io_utils import BlockedArrayWriter
from numcodecs.abc import Codec

import aind_exaspim_pipeline_utils
from xarray_multiscale.reducers import WindowedReducer

import multiprocessing
from typing import Tuple
import time
import psutil
import zarr
import re
from numcodecs import Blosc
import dask
import dask.array
from dask.distributed import Client
from aind_data_transfer.transformations import ome_zarr
from aind_data_transfer.util import chunk_utils, io_utils
from aind_exaspim_pipeline_utils import exaspim_manifest
from aind_exaspim_pipeline_utils.exaspim_manifest import (
    write_result_metadata,
    write_result_manifest,
    append_metadata_to_manifest,
)


def fmt_uri(uri: str) -> str:
    """Format location paths by making slash usage consistent.

    All multiple occurrence internal slashes are replaced to single ones.

    Local paths can be relative or absolute, trailing slash will be added if missing.

    S3 references formatted to pattern ``s3://bucket/path/``.

    Parameters
    ----------
    uri: `str`
        The resource URI.

    Returns
    -------
    r: `str`
      Formatted uri, either a local file system path or an s3:// uri.
    """
    if uri.startswith("s3:") or uri.startswith("S3:"):
        s = "/".join(("", *uri[3:].split("/"), ""))
        return "s3:/" + re.sub(r"/{2,}", "/", s)
    else:
        s = "/".join((*uri.split("/"), ""))
        return re.sub(r"/{2,}", "/", s)


def downsample_and_store(
    arr: dask.array.Array,
    group: zarr.Group,
    n_lvls: int,
    scale_factors: Tuple,
    block_shape: Tuple,
    compressor: Codec = None,
    reducer: WindowedReducer = xarray_multiscale.reducers.windowed_mean,
    fromLevel: int = 1,
) -> list:  # pragma: no cover
    """
    Progressively downsample the input array and store the results as separate arrays in a Zarr group.

    Parameters
    ----------
    arr : da.Array
        The full-resolution Dask array.
    group : zarr.Group
        The output Zarr group.
    n_lvls : int
        The number of pyramid levels.
    scale_factors : Tuple
        The scale factors for downsampling along each dimension.
    block_shape : Tuple
        The shape of blocks to use for partitioning the array.
    compressor : numcodecs.abc.Codec, optional
        The compression codec to use for the output Zarr array. Default is Blosc with "zstd"
        method and compression level 1.
    fromLevel : int
        The first downscaled level to write. `arr` must represent fromLevel - 1. Defaults to 1.
    """

    for arr_index in range(fromLevel, n_lvls):
        LOGGER.info("Creating downsampled level %d in dask.", arr_index)
        first_mipmap = _get_first_mipmap_level(arr, scale_factors, reducer)

        LOGGER.info("Creating dataset for level %d.", arr_index)
        ds = group.create_dataset(
            str(arr_index),
            shape=first_mipmap.shape,
            chunks=first_mipmap.chunksize,
            dtype=first_mipmap.dtype,
            compressor=compressor,
            dimension_separator="/",
            overwrite=True,
        )
        LOGGER.info("Storing downsampled level %d.", arr_index)
        BlockedArrayWriter.store(first_mipmap, ds, block_shape)

        arr = dask.array.from_array(ds, chunks=ds.chunks)


def run_n5tozarr(
    input_uri: str,
    output_uri: str,
    voxel_sizes_zyx: Tuple[float, float, float],
):  # pragma: no cover
    """Run initial conversion and 4 layers of downscaling.

    All output arrays will be 5D, chunked as (1, 1, 128, 128, 128),
    downscaling factor is (1, 1, 2, 2, 2).

    Parameters
    ----------
    input_uri: `str`
        Input uri or path on the local filesystem.
    output_uri: `str`
        Output uri or on the local filesystem.
    voxel_sizes_zyx: tuple of `float`
        Voxel size in microns in the input (full resolution) dataset
    """
    LOGGER.debug("Initialize source N5 store")
    n5s = zarr.n5.N5FSStore(input_uri)
    zg = zarr.open(store=n5s, mode="r")
    LOGGER.debug("Initialize dask array from N5 source")
    arr = dask.array.from_array(zg["s0"], chunks=zg["s0"].chunks)
    arr = chunk_utils.ensure_array_5d(arr)
    LOGGER.debug("Re-chunk dask array to desired output chunk size.")
    arr = arr.rechunk((1, 1, 256, 256, 256))

    LOGGER.info(f"Input array: {arr}")
    LOGGER.info(f"Input array size: {arr.nbytes / 2 ** 20} MiB")

    LOGGER.debug("Initialize target Zarr store")
    output_path = output_uri
    group = zarr.open_group(output_path, mode="a")

    scale_factors = (2, 2, 2)
    scale_factors = chunk_utils.ensure_shape_5d(scale_factors)

    n_levels = 1
    compressor = Blosc(cname="zstd", clevel=1)

    block_shape = chunk_utils.ensure_shape_5d(
        io_utils.BlockedArrayWriter.get_block_shape(arr, target_size_mb=1048576)  # 8k**3 blocks of 16bit data
    )
    LOGGER.info(f"Calculation block shape: {block_shape}")

    # Actual Processing
    ome_zarr.write_ome_ngff_metadata(
        group,
        arr,
        output_uri,
        n_levels,
        scale_factors[2:],
        voxel_sizes_zyx,
        origin=None,
    )

    t0 = time.time()
    LOGGER.info("Starting initial N5 -> Zarr copy.")
    ome_zarr.store_array(arr, group, "0", block_shape, compressor)
    write_time = time.time() - t0
    LOGGER.info(f"Finished writing tile. Took {write_time}s.")


def run_zarr_multiscale(
    input_uri: str, output_uri: str, voxel_sizes_zyx: Tuple[float, float, float], fromLevel: int = 1
):  # pragma: no cover
    """Run downscaling on an existing 0 level zarr.

    All output arrays will be 5D, chunked as (1, 1, 128, 128, 128),
    downscaling factor is (1, 1, 2, 2, 2).

    Parameters
    ----------
    input_uri: `str`
        Input s3 uri path or path on the local filesystem.
    output_uri: `str`
        Output s3 uri path or path on the local filesystem. Can be the same as `input_uri`.
    voxel_sizes_zyx: tuple of `float`
        Voxel size in microns in the input (full resolution) dataset
    """
    LOGGER.debug("Initialize source Zarr store")
    zg = zarr.open_group(input_uri, mode="r")
    LOGGER.info("Get dask array from Zarr source for full resolution")
    arrZero = dask.array.from_array(
        zg["0"], chunks=zg["0"].chunks
    )  # For metadata writing we need the full resolution shape
    arrZero = chunk_utils.ensure_array_5d(arrZero)
    arrZero = arrZero.rechunk((1, 1, 256, 256, 256))

    LOGGER.info(f"Full resolution array: {arrZero}")
    LOGGER.info(f"Full resolution input array size: {arrZero.nbytes / 2 ** 20} MiB")

    LOGGER.debug("Initialize target Zarr store")
    group = zarr.open_group(output_uri, mode="a")

    scale_factors = (2, 2, 2)
    scale_factors = chunk_utils.ensure_shape_5d(scale_factors)

    n_levels = 8
    compressor = Blosc(cname="zstd", clevel=1)

    # Actual Processing
    ome_zarr.write_ome_ngff_metadata(
        group,
        arrZero,
        output_uri,
        n_levels,
        scale_factors[2:],
        voxel_sizes_zyx,
        origin=None,
    )

    if fromLevel > 1:
        prevLevel = str(fromLevel - 1)
        LOGGER.info("Initialize dask source array from Zarr source level %s", prevLevel)
        arr = dask.array.from_array(zg[prevLevel], chunks=zg[prevLevel].chunks)
        arr = chunk_utils.ensure_array_5d(arr)
        arr = arr.rechunk((1, 1, 256, 256, 256))
    else:
        arr = arrZero

    del arrZero  # Can be garbage collected if different from arr

    block_shape = chunk_utils.ensure_shape_5d(
        io_utils.BlockedArrayWriter.get_block_shape(arr, target_size_mb=1048576)  # 8k**3 blocks of 16bit data
    )
    LOGGER.info(f"Calculation block shape: {block_shape}")

    t0 = time.time()
    LOGGER.info("Starting N5 -> downsampled Zarr level copies.")
    downsample_and_store(arr, group, n_levels, scale_factors, block_shape, compressor, fromLevel=fromLevel)
    write_time = time.time() - t0

    LOGGER.info("Finished writing tile. Took %d s.", write_time)


def get_worker_memory(n_worker):  # pragma: no cover
    """Determine the per-worker memory"""
    total = psutil.virtual_memory().total
    GByte = 1024 * 1024 * 1024
    LOGGER.info("Total physical memory: %.1f GiB", total / GByte)
    wmem = total - 24 * GByte  # Reserve for scheduler
    perworker = wmem // n_worker
    if wmem < 0 or perworker < 2 * GByte:
        raise RuntimeError("Not enough memory for 24 GiB for scheduler and at least 2 GiB per worker")
    LOGGER.info("Set aside 24 GiB for scheduler, %.1f GiB per worker process", perworker / GByte)
    return perworker


def set_worker_logging(level: int = logging.DEBUG):  # pragma: no cover
    """Configure logger levels on a worker."""
    logging.getLogger().setLevel(level)
    logging.getLogger("distributed").setLevel(level)
    # Do not want to see network debug stuff
    if level > logging.INFO:
        infolevel = level
    else:
        infolevel = logging.INFO
    logging.getLogger("boto3").setLevel(infolevel)
    logging.getLogger("botocore").setLevel(infolevel)
    logging.getLogger("s3fs").setLevel(infolevel)
    logging.getLogger("urllib3").setLevel(infolevel)


def config_client_logging(level: int = logging.DEBUG):  # pragma: no cover
    """Configure logging on a worker or the client."""
    # config = {
    #     "version": 1,
    #     "handlers": {
    #         "console": {
    #             "class": "logging.StreamHandler",
    #             "formatter": "default",
    #             "level": logging.getLevelName(level),
    #         }
    #     },
    #     "formatters": {
    #         "default": {
    #             # "worker" field was referenced in `Client.forward_logging` documentation but does not work:
    #             # "format": "%(asctime)s %(levelname)-8s [%(process)d %(worker)s] %(name)-15s %(message)s",
    #             "format": "%(asctime)s [%(process)d] %(levelname)s %(name)s %(message)s",
    #             "datefmt": "%Y-%m-%d %H:%M:%S",
    #         }
    #     },
    #     "root": {"handlers": ["console"]},
    # }
    R = logging.getLogger()
    for hdlr in R.handlers[:]:  # remove all old handlers
        R.removeHandler(hdlr)

    h1 = logging.StreamHandler(sys.stderr)
    h1.setFormatter(logging.Formatter("%(asctime)s [%(process)d] %(levelname)s %(name)s %(message)s"))
    logging.getLogger().setLevel(level)
    logging.getLogger().addHandler(h1)
    # logging.config.dictConfig(config)
    # logging.getLogger().setLevel(level)
    # Do not want to see network debug stuff
    if level > logging.INFO:
        infolevel = level
    else:
        infolevel = logging.INFO
    logging.getLogger("distributed.worker").setLevel(infolevel)
    logging.getLogger("distributed").setLevel(infolevel)
    logging.getLogger("boto3").setLevel(infolevel)
    logging.getLogger("botocore").setLevel(infolevel)
    logging.getLogger("s3fs").setLevel(infolevel)
    logging.getLogger("urllib3").setLevel(infolevel)


def n5tozarr_da_converter():  # pragma: no cover
    """Main entry point to n5tozarr converter task."""
    config_client_logging(logging.INFO)
    global LOGGER
    LOGGER = logging.getLogger("n5tozarr")
    LOGGER.setLevel(logging.INFO)

    capsule_manifest = exaspim_manifest.get_capsule_manifest()
    config = capsule_manifest.processing_pipeline.n5_to_zarr.dict()
    if config is None:
        raise ValueError("Manifest does not contain configuration for n5tozarr processing")

    config["input_uri"] = fmt_uri(config["input_uri"])
    config["output_uri"] = fmt_uri(config["output_uri"])
    n_cpu = multiprocessing.cpu_count()

    config["capsule"] = dict(version="n5tozarr_0.1.0", n_cpu=n_cpu)  # TBD: obtain version from CO environment

    # Create initial metadata in case the run crashes
    meta = get_n5tozarr_metadata(config)
    write_result_metadata(meta)

    # Start dask cluster
    LOGGER.info("Starting local Dask cluster with %d processes and 2 threads per process.", n_cpu)
    dask.config.set(
        {
            "distributed.worker.memory.spill": False,  # Do not spill to /tmp space in a capsule
            "distributed.worker.memory.target": False,  # Do not spill to /tmp space in a capsule
            # If a worker gets paused, after the recent-to-old wait time, we hope
            # that it will cross the termination limit and eventually gets restarted
            "distributed.worker.memory.terminate": 0.95,
            # Pause and wait for task finish then GC and trimming:
            "distributed.worker.memory.pause": 0.94,
            # Do not receive data from other workers:
            "distributed.worker.memory.rebalance.recipient-max": 0.05,
            # Should be longer than typical task runtime:
            "distributed.worker.memory.recent-to-old-time": "300s",
        }
    )
    client = Client(
        n_workers=n_cpu,
        threads_per_worker=2,
        memory_limit=get_worker_memory(n_cpu),
        processes=True,
        silence_logs=logging.INFO,
    )
    client.run(set_worker_logging, logging.INFO)
    client.forward_logging()
    # Run jobs
    run_n5tozarr(config["input_uri"], config["output_uri"], config["voxel_size_zyx"])
    # Update metadata to show that we've finished properly
    set_metadata_done(meta)
    write_result_metadata(meta)
    append_metadata_to_manifest(capsule_manifest, meta)
    write_result_manifest(capsule_manifest)
    # Close down
    LOGGER.info("Sleep 120s to get workers into an idle state.")
    time.sleep(120)  # leave time for workers to get into an idle state before shutting down
    LOGGER.info("Closing cluster.")
    client.close(180)  # leave time for workers to exit
    LOGGER.info("Done.")


def get_zarr_multiscale_metadata(config: dict):  # pragma: no cover
    """Initiate metadata instance with current timestamp and configuration."""
    t = datetime.datetime.now()
    dp = DataProcess(
        name=ProcessName.FILE_CONVERSION,
        version=config["capsule"]["version"],
        start_date_time=t,
        end_date_time=t,
        input_location=config["input_uri"],
        output_location=config["output_uri"],
        code_url="https://github.com/AllenNeuralDynamics/aind-exaSPIM-pipeline-utils",
        code_version=aind_exaspim_pipeline_utils.__version__,
        parameters=config,
        outputs=None,
        notes="IN PROGRESS",
    )
    return dp


def get_n5tozarr_metadata(config: dict):  # pragma: no cover
    """Initiate metadata instance with current timestamp and configuration."""
    t = datetime.datetime.now()
    dp = DataProcess(
        name=ProcessName.FILE_CONVERSION,
        version=config["capsule"]["version"],
        start_date_time=t,
        end_date_time=t,
        input_location=config["input_uri"],
        output_location=config["output_uri"],
        code_url="https://github.com/AllenNeuralDynamics/aind-exaSPIM-pipeline-utils",
        code_version=aind_exaspim_pipeline_utils.__version__,
        parameters=config,
        outputs=None,
        notes="IN PROGRESS",
    )
    return dp


def set_metadata_done(meta: DataProcess) -> None:  # pragma: no cover
    """Update end timestamp and set metadata note to ``DONE``.

    Parameters
    ----------
    meta: DataProcess
      Capsule metadata instance.
    """
    t = datetime.datetime.now()
    meta.end_date_time = t
    meta.notes = "DONE"


def zarr_multiscale_converter():  # pragma: no cover
    """Main entry point for zarr downscaling task."""
    config_client_logging(logging.INFO)
    global LOGGER
    LOGGER = logging.getLogger("zarr_mscale")
    LOGGER.setLevel(logging.INFO)

    capsule_manifest = exaspim_manifest.get_capsule_manifest()
    config = capsule_manifest.processing_pipeline.zarr_multiscale.dict()
    if config is None:
        raise ValueError("Manifest does not contain configuration for zarr_multiscale processing")

    # Add dynamic entries to config
    config["input_uri"] = fmt_uri(config["input_uri"])
    if config["output_uri"] is None:
        config["output_uri"] = config["input_uri"]
    else:
        config["output_uri"] = fmt_uri(config["output_uri"])
    n_cpu = multiprocessing.cpu_count()
    config["capsule"] = dict(
        version="zarr_multiscale_0.1.0", n_cpu=n_cpu
    )  # TBD: obtain version from CO environment

    # Create initial metadata in case the run crashes
    meta = get_zarr_multiscale_metadata(config)
    write_result_metadata(meta)

    # Start dask cluster
    LOGGER.info("Starting local Dask cluster with %d processes and 2 threads per process.", n_cpu)
    dask.config.set(
        {
            "distributed.worker.memory.spill": False,  # Do not spill to /tmp space in a capsule
            "distributed.worker.memory.target": False,  # Do not spill to /tmp space in a capsule
            # If a worker gets paused, after the recent-to-old wait time, we hope
            # that it will cross the termination limit and eventually gets restarted
            "distributed.worker.memory.terminate": 0.95,
            # Pause and wait for task finish then GC and trimming:
            "distributed.worker.memory.pause": 0.94,
            # Do not receive data from other workers:
            "distributed.worker.memory.rebalance.recipient-max": 0.05,
            # Should be longer than typical task runtime:
            "distributed.worker.memory.recent-to-old-time": "300s",
        }
    )
    client = Client(
        n_workers=n_cpu,
        threads_per_worker=2,
        memory_limit=get_worker_memory(n_cpu),
        processes=True,
        silence_logs=logging.INFO,
    )
    client.run(set_worker_logging, logging.INFO)
    client.forward_logging()
    # Run jobs
    run_zarr_multiscale(config["input_uri"], config["output_uri"], config["voxel_size_zyx"])
    # Update metadata to show that we've finished properly
    set_metadata_done(meta)
    write_result_metadata(meta)
    append_metadata_to_manifest(capsule_manifest, meta)
    write_result_manifest(capsule_manifest)
    # Close down
    LOGGER.info("Sleep 120s to get workers into an idle state.")
    time.sleep(120)  # leave time for workers to get into an idle state before shutting down
    LOGGER.info("Closing cluster.")
    client.close(180)  # leave time for workers to exit
    LOGGER.info("Done.")


if __name__ == "__main__":
    n5tozarr_da_converter()



"""exaSPIM pipeline utilites top level namespace definition
"""
from .imagej_macros import ImagejMacros
from .imagej_wrapper import main
from .n5tozarr.n5tozarr_da import n5tozarr_da_converter, zarr_multiscale_converter
from .exaspim_manifest import create_example_manifest

__all__ = [
    "ImagejMacros",
    "main",
    "n5tozarr_da_converter",
    "create_example_manifest",
    "zarr_multiscale_converter",
]

__version__ = "0.4.3"



"""Tests for n5tozarr_da_convert functions."""
import unittest
from aind_exaspim_pipeline_utils.n5tozarr.n5tozarr_da import fmt_uri


class TestN5toZarr(unittest.TestCase):
    """N5toZarr tests"""

    def test_uri_formatting(self):
        """Uri string formatter test case"""
        s = fmt_uri("s3:/test_bucket/zarr_root/dataset_name")
        self.assertEqual(s, "s3://test_bucket/zarr_root/dataset_name/")

        s = fmt_uri("/zarr_root/dataset_name")
        self.assertEqual(s, "/zarr_root/dataset_name/")



"""Configuration file for the Sphinx documentation builder."""
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

from datetime import date

# -- Path Setup --------------------------------------------------------------
from os.path import abspath, dirname
from pathlib import Path

from exaspim_pipeline_utils import __version__ as package_version

INSTITUTE_NAME = "Allen Institute for Neural Dynamics"

current_year = date.today().year

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = Path(dirname(dirname(dirname(abspath(__file__))))).name
copyright = f"{current_year}, {INSTITUTE_NAME}"
author = INSTITUTE_NAME
release = package_version

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.duration",
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
]
templates_path = ["_templates"]
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_static_path = ["_static"]
html_favicon = "_static/favicon.ico"
html_theme_options = {
    "light_logo": "light-logo.svg",
    "dark_logo": "dark-logo.svg",
}

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = False

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
html_show_copyright = False



"""Tests for ImageJ wrapper functions and macro creator class"""
import contextlib
import io
import logging
import unittest
from unittest import mock

import argschema

from aind_exaspim_pipeline_utils.imagej_macros import ImagejMacros
from aind_exaspim_pipeline_utils.imagej_wrapper import (
    ImageJWrapperSchema,
    get_auto_parameters,
    wrapper_cmd_run,
)


class TestWrapperFunctions(unittest.TestCase):
    """Tests for indivicual functions"""

    @mock.patch("os.cpu_count")
    @mock.patch("psutil.virtual_memory")
    def testAutoParameters(self, mock_psutil_virtual_memory, mock_os_count):
        """Test for mem and cpu detection"""
        mock_os_count.return_value = 4
        mock_total = mock.Mock(total=128 * 1024 * 1024 * 1024)
        mock_psutil_virtual_memory.return_value = mock_total

        args = {"session_id": "test_session_123"}
        d = get_auto_parameters(args)

        self.assertIn("process_xml", d)
        self.assertEqual(d["auto_ncpu"], 4)
        self.assertEqual(d["auto_memgb"], 128 - 12)

    @mock.patch("subprocess.Popen")
    @mock.patch("selectors.DefaultSelector")
    def testCmdWrapper(self, mock_DefaultSelector, mock_subprocess_popen):
        """Tests for the cmd wrapper"""
        mock_selector = mock.Mock()
        mock_selector.configure_mock(**{"register.return_value": None, "close.return_value": None})
        mock_DefaultSelector.return_value = mock_selector
        mock_std = mock.Mock()
        mock_std.configure_mock(**{"close.return_value": None, "read.return_value": b"text"})
        mock_popen = mock.Mock(stdout=mock_std, stderr=mock_std)
        mock_popen.configure_mock(**{"poll.return_value": 1, "wait.return_value": 0})
        mock_subprocess_popen.return_value = mock_popen

        s_out = io.StringIO()
        s_err = io.StringIO()
        with contextlib.redirect_stdout(s_out), contextlib.redirect_stderr(s_err):
            r = wrapper_cmd_run("test_cmd", logging.getLogger())
        self.assertTrue("text" in s_err.getvalue())
        self.assertTrue("text" in s_out.getvalue())
        self.assertEqual(r, 0)
        mock_std.read.assert_called()
        mock_std.close.assert_called()
        mock_selector.close.assert_called()


class TestMacros(unittest.TestCase):
    """Test case for ImagejMacros"""

    def setUp(self):
        """Set up ArgSchemaParser.args"""
        example_params_default = {
            "session_id": "2023-02-22",
            "memgb": 55,
            "parallel": 8,
            "dataset_xml": "test_dataset.xml",
            "do_detection": True,
            "ip_detection_params": {
                "downsample": 8,
                "bead_choice": "sample_small",  # Usual choice for the beads
                "set_minimum_maximum": True,
                "maximum_number_of_detections": 100000,
                "ip_limitation_choice": "around_median",
            },
            "do_registrations": True,
            "ip_registrations_params": [
                {
                    "transformation_choice": "translation",
                    "compare_views_choice": "overlapping_views",
                    "interest_point_inclusion_choice": "overlapping_ips",
                    "fix_views_choice": "first_fixed",
                    "fixed_tile_ids": [
                        10,
                    ],
                    "map_back_views_choice": "no_mapback",
                    "map_back_reference_view": 5,
                    "do_regularize": True,
                    "regularize_with_choice": "rigid",
                }
            ],
        }
        parser = argschema.ArgSchemaParser(
            schema_type=ImageJWrapperSchema, input_data=example_params_default, args=[]
        )
        self.args = parser.args

    def testMacroIPDet(self):
        """Test IP Detection macro"""

        det_params = dict(self.args["ip_detection_params"])
        det_params["process_xml"] = self.args["dataset_xml"]
        det_params["parallel"] = self.args["parallel"]

        m = ImagejMacros.get_macro_ip_det(det_params)
        self.assertRegex(m, "downsample_z=8x")
        self.assertNotRegex(m, "sigma=")

        det_params["bead_choice"] = "manual"
        m = ImagejMacros.get_macro_ip_det(det_params)
        self.assertRegex(m, "sigma=")
        print(m)

    def testMacroIPReg(self):
        """Test IP Registration macro"""
        reg_params = dict(self.args["ip_registrations_params"][0])
        reg_params["process_xml"] = self.args["dataset_xml"]
        reg_params["parallel"] = self.args["parallel"]
        m = ImagejMacros.get_macro_ip_reg(reg_params)
        self.assertRegex(m, "select=test_dataset.xml")
        self.assertNotRegex(m, "viewsetupid_")

        reg_params["fix_views_choice"] = "select_fixed"
        m = ImagejMacros.get_macro_ip_reg(reg_params)
        self.assertRegex(m, "viewsetupid_10_timepoint_0")

        reg_params["map_back_views_choice"] = "selected_translation"
        m = ImagejMacros.get_macro_ip_reg(reg_params)
        self.assertRegex(m, "ViewSetupId:5")



"""ImageJ Macro creator module."""
from typing import Any, Dict


class ImagejMacros:
    """Generate imageJ macros from template strings by substituting parameter values.

    This class does not check substitution values and expects that all values are valid.
    """

    # Fiji macro allows new line at , separated arguments
    # strings can be added " " + " "
    # within strings, [ ] can be used for spaces, for file names? TBC

    MACRO_IP_DET = """
run("Memory & Threads...", "parallel={parallel:d}");
run("Detect Interest Points for Registration",
"select={process_xml} process_angle=[All angles] process_channel=[All channels] " +
"process_illumination=[All illuminations] process_tile=[All tiles] process_timepoint=[All Timepoints] " +
"type_of_interest_point_detection=Difference-of-Gaussian label_interest_points=beads " +
"subpixel_localization=[3-dimensional quadratic fit] " +
"interest_point_specification=[{bead_choice}] " +
"downsample_xy=[Match Z Resolution (less downsampling)] downsample_z={downsample}x{manual_bead_choice} " +
"compute_on=[CPU (Java)]{manual_minimum_maximum}{limit_amount_of_detections}");
    """

    MAP_BEAD_CHOICE = {
        "very_weak_small": "Very weak & small (beads)",
        "weak_small": "Weak & small (beads)",
        "sample_small": "Comparable to Sample & small (beads)",
        "strong_small": "Strong & small (beads)",
        "manual": "Advanced ...",
    }

    MAP_IP_LIMITATION_CHOICE = {
        "brightest": "Brightest",
        "around_median": "Around median (of those above threshold)",
        "weakest": "Weakest (above threshold)",
    }

    MACRO_IP_REG = """
run("Memory & Threads...", "parallel={parallel:d}");
run("Register Dataset based on Interest Points",
"select={process_xml} process_angle=[All angles] process_channel=[All channels] " +
"process_illumination=[All illuminations] process_tile=[All tiles] " +
"process_timepoint=[All Timepoints] " +
"registration_algorithm=[Precise descriptor-based (translation invariant)] " +
"registration_in_between_views=[{compare_views}] " +
"interest_point_inclusion=[{interest_point_inclusion}] " +
"interest_points=beads fix_views=[{fix_views}] " +
"map_back_views=[{map_back_views}] " +
"transformation={transformation}{regularization} " +
"number_of_neighbors=3 redundancy=1 significance=3 " +
"allowed_error_for_ransac=5 ransac_iterations=Normal{fixed_viewsetupids}{select_reference_views}");
"""

    TEMPLATE_REGULARIZE = """ regularize_model model_to_regularize_with={regularize_with} lambda=0.10"""

    MAP_COMPARE_VIEWS = {
        "all_views": "Compare all views against each other",
        "overlapping_views": "Only compare overlapping views (according to current transformations)",
    }
    MAP_INTEREST_POINT_INCLUSION = {
        "overlapping_ips": "Only compare interest points that overlap "
        + "between views (according to current transformations)",
        "all_ips": "Compare all interest point of overlapping views",
    }

    MAP_FIX_VIEWS = {
        "no_fixed": "Do not fix views",
        "first_fixed": "Fix first view",
        "select_fixed": "Select fixed view",
    }

    MAP_MAP_BACK_VIEWS = {
        "no_mapback": "Do not map back (use this if views are fixed)",
        "first_translation": "Map back to first view using translation model",
        "first_rigid": "Map back to first view using rigid model",
        "selected_translation": "Map back to user defined view using translation model",
        "selected_rigid": "Map back to user defined view using rigid model",
    }

    MAP_TRANSFORMATION = {"translation": "Translation", "rigid": "Rigid", "affine": "Affine"}

    MAP_REGULARIZATION = {
        "identity": "Identity",
        "translation": "Translation",
        "rigid": "Rigid",
        "affine": "Affine",
    }

    @staticmethod
    def get_macro_ip_det(P: Dict[str, Any]) -> str:
        """Get a parameter formatted IP detection macro.

        Parameters
        ----------
        P : `dict`
          Parameter dictionary for macro formatting.
        """
        fparams = dict(P)
        fparams["manual_bead_choice"] = ""
        if fparams["bead_choice"] == "manual":
            fparams["find_minima"] = " find_minima" if P["find_minima"] else ""
            fparams["find_maxima"] = " find_maxima" if P["find_maxima"] else ""
            fparams[
                "manual_bead_choice"
            ] = " sigma={sigma:.5f} threshold={threshold:.5f}{find_minima}{find_maxima}".format(**fparams)
        fparams["bead_choice"] = ImagejMacros.MAP_BEAD_CHOICE[P["bead_choice"]]

        fparams["manual_minimum_maximum"] = ""
        if fparams["set_minimum_maximum"]:
            fparams["manual_minimum_maximum"] = (
                " set_minimal_and_maximal_intensity minimal_intensity={minimal_intensity:.1f}"
                " maximal_intensity={maximal_intensity:.1f}"
            ).format(**fparams)

        fparams["limit_amount_of_detections"] = ""
        if P["maximum_number_of_detections"] > 0:
            fparams["limit_amount_of_detections"] = " limit_amount_of_detections maximum_number={:d}".format(
                P["maximum_number_of_detections"]
            ) + " type_of_detections_to_use=[{}]".format(
                ImagejMacros.MAP_IP_LIMITATION_CHOICE[P["ip_limitation_choice"]]
            )
        return ImagejMacros.MACRO_IP_DET.format(**fparams)

    @staticmethod
    def get_macro_ip_reg(P: Dict[str, Any]) -> str:
        """Get a parameter formatted IP registration macro.

        Parameters
        ----------
        P : `dict`
          Parameter dictionary for macro template formatting.
        """
        fparams = dict(P)
        fparams["compare_views"] = ImagejMacros.MAP_COMPARE_VIEWS[P["compare_views_choice"]]
        fparams["interest_point_inclusion"] = ImagejMacros.MAP_INTEREST_POINT_INCLUSION[
            P["interest_point_inclusion_choice"]
        ]
        fparams["fix_views"] = ImagejMacros.MAP_FIX_VIEWS[P["fix_views_choice"]]
        fparams["transformation"] = ImagejMacros.MAP_TRANSFORMATION[P["transformation_choice"]]
        fparams["map_back_views"] = ImagejMacros.MAP_MAP_BACK_VIEWS[P["map_back_views_choice"]]
        fparams["regularization"] = ""
        fparams["fixed_viewsetupids"] = ""
        if P["do_regularize"]:
            fparams["regularization"] = ImagejMacros.TEMPLATE_REGULARIZE.format(
                regularize_with=ImagejMacros.MAP_REGULARIZATION[P["regularize_with_choice"]]
            )
        if P["fix_views_choice"] == "select_fixed":
            fparams["fixed_viewsetupids"] = "".join(
                f" viewsetupid_{tile_id:d}_timepoint_0" for tile_id in P["fixed_tile_ids"]
            )
        fparams["select_reference_views"] = ""
        if P["map_back_views_choice"] in ("selected_translation", "selected_rigid"):
            fparams[
                "select_reference_views"
            ] = " select_reference_views=[ViewSetupId:{:d} Timepoint:0]".format(P["map_back_reference_view"])
        return ImagejMacros.MACRO_IP_REG.format(**fparams)
